local CLUSTERMETHODS = {};

local function DEBUG(msg) 
   DEFAULT_CHAT_FRAME:AddMessage("SRUC: " .. msg);
end

function CLUSTERMETHODS:ApplyLocationOptions(opts)
   local parent = self:GetParent();
   local pHeight = parent:GetHeight();
   local pWidth = parent:GetWidth();
   local locX = (opts.locX or 0) / 2.0;
   local locY = (opts.locY or 0) / 2.0;
   local mHeight = self:GetHeight();
   local scale = (pHeight / mHeight) * (opts.scale or 0.5);
   self:SetScale(scale);
   self:ClearAllPoints();
   self:SetPoint("CENTER", parent:GetName(), "CENTER", 
		 (pWidth * locX) / scale,
		 (pHeight * locY) / scale);
end

function CLUSTERMETHODS:ApplyUnitConfigMap(configSet, noSticky)
   local sType = type(unitSpec);
   
   for _,frame in ipairs(self.unitFrames) do
      local unit = frame.unit;
      local config = configSet[unit];
      if (config) then
	 frame:ApplyConfig(config, notSticky);
      end
   end
end

function CLUSTERMETHODS:RefreshVisualStates(visState, force)
   for _,frame in ipairs(self.unitFrames) do
      frame:RefreshVisualState(visState, force);
   end
end

function CLUSTERMETHODS:PreviewVisualStates(unitSpec, visState, fallbackState, force)
   local sType = type(unitSpec);
   
   for _,frame in ipairs(self.unitFrames) do
      local doApply;
      if (sType == "table") then
	 doApply = unitSpec[frame.unit];
      elseif (sType == "string") then
	 doApply = (unitSpec == frame.unit);
      elseif (sType == "function") then
	 doApply = unitSpec(frame.unit);
      end

      if (doApply) then
	 frame:SetVisualState(visState, true, force);
      elseif (fallbackState) then
	 frame:SetVisualState(fallbackState, true, force);
      end
   end
end

function CLUSTERMETHODS:RestoreVisualStates(force)
   for _,frame in ipairs(self.unitFrames) do
      frame:SetVisualState(nil, true, force);
   end
end

function CLUSTERMETHODS:ApplyLayout(layout)
   local layoutManager = StatRingsLayout_CreateLayoutManager();

   this.layoutManager = layoutManager;
   this.layout = layout;

   layoutManager:ProcessLayout(layout);
   for _,frame in ipairs(self.unitFrames) do
      local unit = frame.unit;
      local anchor = layoutManager:GetAnchor(unit);
      if (type(anchor) == "string") then
	 DEBUG("Failed to get location for " .. unit .. ": " .. anchor);
	 frame:Hide();
      else
	 frame:ApplyLayoutAnchor(anchor, this);
      end
   end
end

function StatRingsCluster_OnLoad(this, unitFrames)
   if (not unitFrames) then
      message("StatRingsCluster_OnLoad called without any unitFrames");
   end
   this.unitFrames = unitFrames;

   local newMeta = {};
   local oldMeta = getmetatable(this);
   for k,v in pairs(oldMeta) do newMeta[k] = v; end
   function newMeta.__index(table, index) 
      return oldMeta.__index(table,index) or CLUSTERMETHODS[index];
   end
   setmetatable(this, newMeta);
end
