--STAT_RINGS_TEST_UNIT="player";
--STAT_RINGS_TEST_PET="pet";

local function DEBUG(msg)
   DEFAULT_CHAT_FRAME:AddMessage("SRUF: " .. msg);
end

local METHODS = {};

local RINGLAYOUT_SPLIT = {
   SetupRings =
      function(self, frame)
	 local rotation = frame.rotation or 0;

	 frame.manaFrame:SetScale(1.0);
	 frame.manaFrame:SetReversed(true);
	 frame.manaFrame:SetAngle(self:GetManaAngle(0, rotation));

	 frame.healthFrame:SetScale(1.0);
	 frame.healthFrame:SetReversed(false);
	 frame.healthFrame:SetAngle(self:GetHealthAngle(0, rotation));
      end,

   maxAngle = 120,
   startAngle = 210,
   
   GetHealthAngle = 
      function(self, health, rotation) 
	 return self.maxAngle * health, self.startAngle + rotation
      end,

   GetManaAngle = 
      function(self, mana, rotation) 
	 return self.maxAngle * mana, self.startAngle - rotation; 
      end,
}

-- Make sure ALL of the configuratoin settings are in here
local RINGCONFIG_STANDARD = {
   alphaFull = 0,
   timeFull = 4,
   alphaAlert = 0.7,
   timeAlert = 1,
   healthAlert = 0.5,
   manaUnfade = 1,
   alphaNormal = 0.5,
   timeNormal = 1,
}

function StatRingsUnitFrame_SetRadialPosition(frame, angle, offset, parent)
   if (not parent) then
      parent = frame:GetParent();
      if (parent) then parent = parent:GetName(); end;
   end
   local rad = math.rad(angle);
   frame:SetPoint("CENTER", parent, "CENTER",
		  math.sin(rad) * offset, math.cos(rad) * offset);
end

METHODS.SetRadialPosition = StatRingsUnitFrame_SetRadialPosition;

function METHODS:HandleShow()
   -- If we've been asked to show but are configured not to
   -- then hide again
   if (not self.visShow) then
      self:Hide();
      return;
   end
   -- Otherwise hop into the visual refresh to make sure everything
   -- gets updated
   if (self.fadeGoal) then
      local alpha = self.fadeGoal;
      self.curAlpha = alpha;
      self:SetAlpha(alpha);
   end

   self:RefreshVisualState(self.visualState);
end

function METHODS:SetVisualState(stateData, preview, force)
   local oldVis = self.savedVisualState;
   if (preview) then
      local curVis = self.visualState;

      if (stateData == nil) then
	 oldVis.forceAlpha = true; -- Ensure alpha is restored
	 self.visualState = oldVis;
	 self:RefreshVisualState(oldVis, force);
	 self.savedVisualState = nil;
      else
	 if (stateData ~= curVis) then
	    if (not oldVis) then
	       self.savedVisualState = curVis;
	    end
	 end
	 stateData.forceAlpha = true; -- Ensure alpha is restored
	 self.visualState = stateData;
	 self:RefreshVisualState(stateData, force);
      end
   elseif (stateData) then
      if (oldVis) then
	 -- Override is active
	 self.savedVisualState = stateData;
	 return;
      end
      
      stateData.forceAlpha = true; -- Ensure alpha is restored
      self.visualState = stateData;
      self:RefreshVisualState(stateData, force);
   end
end

function METHODS:RefreshVisualState(stateData, force)
   -- Determine if we care about this visual state update
   if (not stateData) then
      stateData = self.visualState;
   elseif (self.visualState ~= stateData) then
      return;
   end

   --DEBUG("Refreshing visual state for " .. self.unit);

   local show = stateData.show;
   if (force or (show ~= self.visShow)) then
      self.visShow = show;
      if (show) then
	 if (not self:IsVisible()) then
	    stateData.forceAlpha = true;
	 end
	 self:Show();
      else
	 self:Hide();
	 return;
      end
   end

   -- If not visible, exit now and catch up later
   if ((not show) or (not self:IsVisible())) then
      stateData.forceAlpha = true;
      return;
   end

   local health = stateData.health or 0;
   local healthChange;
   if (force or (self.visHealth ~= health)) then
      self.visHealth = health;
      healthChange = true;
   end

   local mana = stateData.mana or 0;
   if (force or (self.visMana ~= mana)) then
      self.visMana = mana;
   else
      mana = nil;
   end
   
   local powerType = stateData.powerType or 0;
   local powerTypeChange;
   if (force or (self.visPowerType ~= powerType)) then
      self.visPowerType = powerType;
      powerTypeChange = true;
   end

   local isGrey = stateData.isGrey;
   local isGreyChange;
   if (force or (self.visIsGrey ~= isGrey)) then
      self.visIsGrey = isGrey;
      isGreyChange = true;
   end

   local alpha = stateData.alpha or 0;
   local alphaSpeed = stateData.alphaSpeed or 0;
   if (alpha and stateData.forceAlpha) then
      alphaSpeed = 0;
      stateData.forceAlpha = nil;
   end
   if (force 
       or (self.visAlpha ~= alpha)
	  or (self.visAlphaSpeed ~= alphaSpeed)) then
      self.visAlpha = alpha;
      self.visAlphaSpeed = alphaSpeed;
   else
      alpha = nil;
      alphaSpeed = nil;
   end

   local name = stateData.name or "";
   if (force or (self.visName ~= name)) then
      self.name = name;
   else
      name = nil;
   end

   if (healthChange or isGreyChange) then
      local r,g,b = 0,1,0;
      if (self.isTarget) then
	 r,g,b = 0.0,0.25,0.0;
      end
      if (self.smoothColor) then
	 if (value < 0.5) then
	    r = (1.0 - value) * 2;
	    g = 1.0;
	 else
	    r = 1.0;
	    g = value * 2;
	 end
      end
      if (isGrey) then
	 local y = r*0.5+g*0.3+b*0.2;
	 r=y; g=y; b=y;
      end
      self.healthFrame:CallTextureMethod("SetVertexColor", r, g, b, 1);
      self.healthFrame:SetAngle(self.ringLayout:GetHealthAngle(health, 
							       self.rotation));
   end
   if (powerTypeChange or isGreyChange) then
      local info = ManaBarColor[powerType] or ManaBarColor[0];
      if (info) then
	 local r,g,b;
	 if (self.isTarget) then
	    r = info.r/4; g=info.g/4; b=info.b/4;
	 else
	    r = info.r;   g=info.g;   b=info.b;
	 end
	 if (isGrey) then
	    local y = r*0.5+g*0.3+b*0.2;
	    r=y; g=y; b=y;
	 end
	 self.manaFrame:CallTextureMethod("SetVertexColor", r, g, b, 1);
      end
   end
   if (mana) then
      self.manaFrame:SetAngle(self.ringLayout:GetManaAngle(mana,
							   self.rotation));
   end
   if (name) then
      if (name == "") then
	 self.nameFrame:Hide();
      elseif (name == true) then
	 self.nameFrame:SetText("<"..self.unit..">");
	 self.nameFrame:Show();
      else
	 self.nameFrame:SetText(name);
	 self.nameFrame:Show();
      end
   end
   if (alpha) then
      if (self.curAlpha ~= alpha) then
	 if (alphaSpeed == 0) then
	    --DEBUG("FadeSet: " .. alpha);
	    self:FadeSet(alpha);
	 elseif (alphaSpeed < 0) then
	    --DEBUG("FadeOut: " .. alpha);
	    self:FadeOut(alpha, -alphaSpeed);
	 else
	    --DEBUG("FadeIn: " .. alpha);
	    self:FadeIn(alpha, alphaSpeed);
	 end
      end
   end
end

function METHODS:UpdateMana(noUpdate)
   local unit = self.unit;

   local mm = UnitManaMax(unit);
   local m = UnitMana(unit);
   self.noMana  = (mm == 0);
   if (mm < 1) then mm = 1 end
   local value = (m/mm);

   local vis = self.eventVisualState;
   if (vis.mana ~= value) then
      vis.mana = value;
      self:CheckFade();
      if (not noUpdate) then
	 self:RefreshVisualState(vis);
      end
   end
end

function METHODS:UpdateManaColor(noUpdate)
   local pType = UnitPowerType(self.unit) or 0;
   self.isRage = (pType == 1) or nil;
   
   local vis = self.eventVisualState;
   if (vis.powerType ~= pType) then
      vis.powerType = pType;
      if (not noUpdate) then
	 self:RefreshVisualState(vis);
      end
   end
end

function METHODS:UpdateHealth(force)
   local hm = UnitHealthMax(self.unit);
   local h = UnitHealth(self.unit);
   if (hm < 1) then hm = 1 end
   local value = h / hm;
   if ((not UnitExists(self.unit)) or (UnitIsDeadOrGhost(self.unit))) then
      value = 0;
   end
   if (value < 0) then
      value = 0;
   elseif (value > 1) then
      value = 1;
   end
      
   local vis = self.eventVisualState;
   if (vis.health ~= value) then
      vis.health = value;
      self:CheckFade();
      if (not noUpdate) then
	 self:RefreshVisualState(vis);
      end
   end
end

function METHODS:UpdateName(noUpdate)
   local name;
   if (not self.hideName) then
      if (UnitExists(self.unit)) then
	 name = UnitName(self.unit);
      end
   end
   
   local vis = self.eventVisualState;
   if (vis.name ~= name) then
      vis.name = name;
      if (not noUpdate) then
	 self:RefreshVisualState(vis);
      end
   end
end

-- Check to see if this frame should be shown
function METHODS:UpdateShown(noUpdate)
   local show = UnitExists(self.unit);
   local vis = self.eventVisualState;
   if (vis.show ~= show) then
      vis.show = show;
      vis.forceAlpha = true;
      self:CheckFade(true);
      self:UpdateName(true);
      if (not noUpdate) then
	 self:RefreshVisualState(vis);
      end
   end
end

local function DoFade()
   if ((not this.fading) or (not this.fadeGoal)) then
      -- DEBUG("Cancelling fade");
      this.fadeStart = nil;
      this.fading = nil;
      this.fadeGoal = nil;
      this:SetScript("OnUpdate", nil);
      return;
   end
   local now = GetTime();
   local elapsed = now - this.fadeStart;
   if (arg1 < elapsed) then
      elapsed = arg1;
   end
   this.fadeStart = now;
   local motion = this.fading * elapsed;
   local newLevel = this.curAlpha + motion;
   if (motion < 0) then
      if (newLevel < this.fadeGoal) then
	 newLevel = this.fadeGoal;
	 this.fading = nil;
      end
   else 
      if (newLevel > this.fadeGoal) then
	 newLevel = this.fadeGoal;
	 this.fading = nil;
      end
   end
   this:SetAlpha(newLevel);
   this.curAlpha = newLevel;
end

local function ActivateFade(frame)
   if (not frame:IsVisible()) then
      frame:SetAlpha(frame.fadeGoal);
      frame.curAlpha = frame.fadeGoal;
      frame.fadeGoal = nil;
      frame.fading = nil;
      return;
   end
   if (not frame.fadeStart) then
      frame.fadeStart = GetTime();
   end
   if (not frame.curAlpha) then
      frame.curAlpha = frame:GetAlpha();
   end
   -- DEBUG("Starting fade " .. frame.fadeStart .. " " .. frame.fading .. " " .. frame.fadeGoal);
   frame:SetScript("OnUpdate", DoFade);
end


function METHODS:FadeSet(level) 
   self:SetAlpha(level);
   self.curAlpha = level;
   self.fading = nil;
end

function METHODS:FadeIn(level, time) 
   if (level > 1) then
      level = 1;
   end
   -- Test if already faded
   if ((self.curAlpha or -1) >= level) then
      return;
   end
   if (time <= 0) then
      time = 0.00001;
   end
   local rate = 1.0 / time;
   local oldFading = self.fading;
   if ((not self.fading) or (rate > self.fading)) then
      self.fading = rate;
   end
   if ((not self.fadeGoal) or (level > self.fadeGoal)) then
      self.fadeGoal = level;
   end

   if (not oldFading) then
      ActivateFade(self);
   end
end

function METHODS:FadeOut(level, time) 
   if (level < 0) then
      level = 0;
   end
   -- Test if already faded
   if ((self.curAlpha or 2) <= level) then
      return;
   end
   if (time <= 0) then
      time = 0.00001;
   end
   local rate = -1.0 / time;
   local oldFading = self.fading;
   if ((not self.fading) or (rate < self.fading)) then
      self.fading = rate;
   end
   if ((not self.fadeGoal) or (level < self.fadeGoal)) then
      self.fadeGoal = level;
   end

   if (not oldFading) then
      ActivateFade(self);
   end
end

function METHODS:GetCurrentFade()
   local vis = self.eventVisualState;
   local fullMana = self.noMana;
   if (self.isRage) then
      if (vis.mana <= (1-self.manaUnfade)) then fullMana = 1; end
   else
      if (vis.mana >= self.manaUnfade) then fullMana = 1; end
   end

   if ((vis.health == 1) and fullMana) then
      if (self.alphaFull < (self.curAlpha or 0)) then
	 return self.alphaFull, -self.timeFull;
      else
	 return self.alphaFull, self.timeFull;
      end
   elseif ((vis.health) and (vis.health < self.healthAlert) 
	   and (not self.isTarget)) then
      if (self.alphaAlert < (self.curAlpha or 0)) then
	 return self.alphaAlert, -self.timeAlert;
      else
	 return self.alphaAlert, self.timeAlert;
      end
   else
      return self.alphaNormal, self.timeNormal;
   end
end
   

function METHODS:CheckFade(noUpdate)
   local vis = self.eventVisualState;
   --DEBUG("CheckingFade " .. ((vis.show and "shown") or "hidden"));
   if (not vis.show) then return; end;
   local newFade, speed = self:GetCurrentFade();
   if (force) then
      vis.alpha = newFade;
      vis.alphaSpeed = 0;
      vis.forceAlpha = true;
   else
      vis.alpha = newFade;
      vis.alphaSpeed = speed;
   end
   if (not noUpdate) then
      self:RefreshVisualState(vis);
   end
end

function METHODS:ApplyConfig(configSet, nonSticky, forceChange)
   local newVal,curVal;
   local anyChange = forceChange;
   for k,v in pairs(RINGCONFIG_STANDARD) do
      newVal = configSet[k] or v;
      curVal = self[k];
      if (curVal ~= newVal) then
	 self[k] = newVal;
	 anyChange = true;
      end
   end
   if (nonSticky) then
      self.currentConfig = nil;
   else
      self.currentConfig = configSet;
   end
   if (anyChange) then
      -- Check visibility, if it didn't change then update health and
      -- mana.
      self:UpdateShown(true);
      self:UpdateHealth(true);
      self:UpdateMana(true);
      self:UpdateManaColor(true);
      self:CheckFade(true);
      self:RefreshVisualState(nil, true);
   end
end

function METHODS:ApplyLayoutAnchor(anchor, parent)
   local parName = (parent or self:GetParent()):GetName();
   
   local scale = anchor.rootScale or 1;
   local x = anchor.x or 0;
   local y = anchor.y or 0;
   self.rotation = anchor.angle or 0;

   self:SetScale(scale);
   self:ClearAllPoints();
   self:SetPoint("CENTER", parName, "CENTER", x/scale, y/scale);
end

function StatRingsUnitFrame_OnLoad(this, unit, hideName)
   local name = this:GetName();
   this.unit = unit;
   this.healthFrame = getglobal(name .. "Health");
   this.manaFrame = getglobal(name .. "Mana");
   this.nameFrame = getglobal(name .. "Name");

   this.hideName = hideName;
   this.rotation = 0;

   this.isTarget = (unit == "target");

   local partyUnit = string.sub(unit, 1, 5) == "party";
   local petUnit = (unit=='pet' or string.sub(unit, -4, -1) == "pet");

   this:RegisterEvent("PLAYER_ENTERING_WORLD");
   this:RegisterEvent("UNIT_AURA");
   if (petUnit) then
      this:RegisterEvent("UNIT_PET");
   end
   if (unit == "target") then
      this:RegisterEvent("PLAYER_TARGET_CHANGED");
   end
   if (partyUnit) then
      this:RegisterEvent("PARTY_MEMBERS_CHANGED");
   end
   this:RegisterEvent("UNIT_NAME_UPDATE");
   this:RegisterEvent("UNIT_MANA");
   this:RegisterEvent("UNIT_MAXMANA");
   this:RegisterEvent("UNIT_FOCUS");
   this:RegisterEvent("UNIT_MAXFOCUS");
   this:RegisterEvent("UNIT_ENERGY");
   this:RegisterEvent("UNIT_MAXENERGY");
   this:RegisterEvent("UNIT_RAGE");
   this:RegisterEvent("UNIT_MAXRAGE");
   this:RegisterEvent("UNIT_HEALTH");
   this:RegisterEvent("UNIT_MAXHEALTH");
   this:RegisterEvent("UNIT_DISPLAYPOWER");

   local newMeta = {};
   local oldMeta = getmetatable(this);
   for k,v in pairs(oldMeta) do newMeta[k] = v; end
   function newMeta.__index(table, index) 
      return oldMeta.__index(table,index) or METHODS[index];
   end
   setmetatable(this, newMeta);
   
   this.manaFrame:SetPoint("CENTER", name, "CENTER", 0, 0);
   this.healthFrame:SetPoint("CENTER", name, "CENTER", 0, 0);

   this.eventVisualState = {
      show = false,
      health = 0,
      mana = 0,
   }
   this.visualState = this.eventVisualState;

   this.ringLayout = RINGLAYOUT_SPLIT;
   this.ringLayout:SetupRings(this);

   -- Apply default configuration but dont link to it
   this:ApplyConfig(RINGCONFIG_STANDARD, true, true);
   this:RefreshVisualState();
end

function StatRingsUnitFrame_ComboOnEvent(this, event)
   this:SetAngle(130 * GetComboPoints() / MAX_COMBO_POINTS);
end

function METHODS:HandleEvent(event, arg1)
   if (event == "PLAYER_ENTERING_WORLD") then
      self:UpdateHealth(true);
      self:UpdateManaColor(true);
      self:UpdateMana(true);
      self:UpdateName(true);
      self:UpdateShown(true);
      self:RefreshVisualState(self.eventVisualState);
      return;
   end

   if (event == "UNIT_COMBAT") then
      DEBUG("UNIT_COMBAT: " .. (arg1 or ''));
      return;
   end

   if (event == "UNIT_PET") then
      if ((arg1 == "player") and (self.unit == "pet")) then
	 self:UpdateShown();
      end
      return;
   end

   if (event == "PLAYER_TARGET_CHANGED") then
      -- No need to check the unit, since this event is only registered
      -- for the target.
      self:UpdateHealth(true);
      self:UpdateManaColor(true);
      self:UpdateMana(true);
      self:UpdateName(true);
      self:UpdateShown(true);
      self:RefreshVisualState(self.eventVisualState);
      self:UpdateShown();
      return;
   end

   if (event == "PARTY_MEMBERS_CHANGED") then
      -- No need to check the unit, since this event is only registered
      -- for party units.
      self:UpdateHealth(true);
      self:UpdateManaColor(true);
      self:UpdateMana(true);
      self:UpdateName(true);
      self:UpdateShown(true);
      self:RefreshVisualState(self.eventVisualState);
      self:UpdateShown();
      return;
   end

   if (arg1 ~= self.unit) then
      return;
   end

   if (event == "UNIT_NAME_UPDATE") then
      if (UnitExists(self.unit)) then
	 self:UpdateName();
      end
      return;
   end

   if (event == "UNIT_DISPLAYPOWER") then
      self:UpdateManaColor();
      self:UpdateMana();
      return;
   end

   if (event == "UNIT_AURA") then
      -- TODO - Avoid repeated updates
      self:UpdateHealth();
      self:UpdateManaColor();
      self:UpdateMana();
      return;
   end

   if ((event == "UNIT_HEALTH") or (event == "UNIT_MAXHEALTH")) then
      self:UpdateHealth();
      return;
   end

   self:UpdateMana();
end
